# Copyright (c) 2004  Daiki Ueno <ueno@unixuser.org>
# You can redistribute it and/or modify it under either the terms of the GPL.

# Ristet is a remake of Tesrit, an old Tetris style game run on MSX.
# Tesrit uses only one column of the grid and blocks can be rotated
# with Left/Right keys.  When the inclination of the current block
# matches the previous one, these blocks are cleared off screen and
# scores are added.

# My best score is 45572 (LEVEL 17).

require 'sdl'
require 'observer'

Colors = [
  [255, 165, 0],		# orange
  [152, 251, 152],		# PaleGreen
  [99, 184, 255],		# SteelBlue1
  [30, 144, 255],		# DodgerBlue1
  [205, 205, 0],		# yellow3
  [205, 0, 205],		# magenta3
  [221, 160, 221],		# plum
  [0, 205, 205],		# cyan3
  [255, 0, 0],			# red
  [225, 228, 181],		# moccasin
  [255, 255, 0],		# yellow
  [255, 215, 0],		# gold
].freeze

DistanceTable = [
  [10, 20, 28, 35, 40, 42, 42, 42, 40, 35, 28, 20],
  [20, 11, 20, 29, 36, 41, 44, 44, 44, 41, 36, 29],
  [28, 20, 12, 22, 32, 40, 45, 48, 49, 48, 45, 36],
  [35, 29, 22, 14, 27, 39, 48, 55, 59, 59, 48, 41],
  [40, 36, 32, 27, 18, 37, 53, 67, 69, 59, 49, 44],
  [42, 41, 40, 39, 37, 33, 69, 74, 67, 55, 48, 44],
  [40, 42, 43, 46, 51, 67, 72, 67, 51, 46, 43, 42],
  [42, 44, 48, 55, 67, 74, 69, 33, 37, 39, 40, 41],
  [40, 44, 49, 59, 69, 67, 53, 37, 18, 27, 32, 36],
  [35, 41, 48, 59, 59, 55, 48, 39, 27, 14, 22, 29],
  [28, 36, 45, 48, 49, 48, 45, 40, 32, 22, 12, 20],
  [20, 29, 36, 41, 44, 44, 44, 41, 36, 29, 20, 11]
].freeze

Roof = 436
ScreenWidth = 480
ScreenHeight = 512
ItemWidth = 72
ItemHeight = 8
ItemBound = 74
ItemCenter = ItemBound / 2
PanelWidth = ItemBound + 20
PanelHeight = ScreenHeight - ItemBound

class Item
  def initialize(y, angle)
    @y, @angle = y, angle
  end
  attr_accessor :y, :angle
end

class Model
  include Observable

  def initialize
    @items = []
    @last_item = Item.new(- ItemCenter, 0)
    @removed_items = []
    @score = 0
    @level = 1
  end
  attr_reader :last_item, :current_item, :removed_items

  def items
    @items.compact
  end

  attr_reader :score, :level

  def create_item(angle)
    @current_item = Item.new(Roof, angle)
  end

  def stopped?
    @current_item.y - @last_item.y <=
      DistanceTable[@last_item.angle][@current_item.angle]
  end

  def fixable?
    @current_item.y < Roof
  end

  def fix_item
    @items.push(Item.new(@last_item.y +
			 DistanceTable[@last_item.angle][@current_item.angle],
			 @current_item.angle))
    @current_item = nil
    (@items.length - 1).downto(1) do |i|
      item = @items[i]
      next unless item
      j = i - 1
      (i - 1).downto(0) do |j|
	break if @items[j]
      end
      next unless @items[j] && @items[j].angle == item.angle
      removed = [i, j]
      (0 .. j - 1).each {|k|
	removed << k if @items[k] && @items[k].angle == item.angle
      }
      @score += @level * (2 ** (removed.length - 1))
      @removed_items = removed.collect {|j| @items[j]}
      changed
      notify_observers
      removed.each {|j| @items[j] = nil}
      @removed_items.clear

      last = Item.new(- ItemCenter, 0)
      @items.each {|pointer|
	next unless pointer
	pointer.y = last.y + DistanceTable[last.angle][pointer.angle]
	last = pointer
      }

      changed
      notify_observers
    end
    @items.compact!
    if @items.empty?
      @last_item = Item.new(- ItemCenter, 0)
    else
      @last_item = @items.last
    end
  end

  def modify_item(y, angle)
    @current_item.y = y
    @current_item.angle = angle
    changed
    notify_observers
  end

  def level_up
    @level += 1
    changed
    notify_observers
  end
end

class View
  def initialize(model, screen, x, y, width, height, font)
    @model, @screen, @x, @y, @width, @height, @font =
      model, screen, x, y, width, height, font
    @margin = ItemBound - ItemWidth
    12.times do |i|
      build_color(i)
      build_image(i)
    end
    @model.add_observer(self)
  end

  def update
    draw_background
    draw_info(@model.score, @model.level)
    draw_items(@model.items, false)
    @model.removed_items.length.times do
      draw_items(@model.removed_items, true)
      @screen.flip
      sleep(0.03)
      draw_items(@model.removed_items, false)
      @screen.flip
      sleep(0.03)
    end
    draw_item(@model.current_item, false) if @model.current_item
    @screen.flip
  end

  private
  def draw_item(item, bright)
    image = Images[item.angle]
    SDL.blit_surface(bright ? image.bright : image.normal, 0, 0, 0, 0,
		     @screen,
		     @x + @width / 2 - ItemCenter,
		     @y + @height - ItemBound - item.y)
  end

  def draw_items(items, bright)
    items.each {|item| draw_item(item, bright)}
  end

  def draw_background
    @screen.fill_rect(0, 0, @screen.w, @screen.h, 0)
    @screen.draw_filled_rect_alpha(@x, @y, @width, @height,
 				   @screen.map_rgb(60, 60, 60), 255)
  end

  def draw_info(score, level)
    ypos = 20
    @font.draw_solid_utf8(@screen, "RISTET", 20, ypos, 255, 255, 255)
    ypos += @font.height
    @font.draw_solid_utf8(@screen, "SCORE: #{score}", 20, ypos, 205, 0, 205)
    ypos += @font.height
    @font.draw_solid_utf8(@screen, "LEVEL: #{level}", 20, ypos, 205, 205, 0)
  end

  def draw_message(message)
    @font.draw_solid_utf8(@screen, message, 20, 200, 205, 205, 205)
  end

  Colors = Array.new(12)
  Images = Array.new(12)

  class Image
    def initialize(normal, bright)
      @normal, @bright = normal, bright
    end
    attr_reader :normal, :bright
  end

  def build_color(n)
    Colors[n] = @screen.map_rgb(*::Colors[n])
  end

  def build_image(n)
    angle = 15 * n

    base_normal = build_base_surface(Colors[n])
    normal = build_empty_surface
    SDL.transform_blit(base_normal, normal, angle, 1, 1,
		       ItemCenter, ItemCenter,
		       ItemCenter, ItemCenter, 0)

    base_bright = build_base_surface(@screen.map_rgb(255, 255, 255))
    bright = build_empty_surface
    SDL.transform_blit(base_bright, bright, angle, 1, 1,
		       ItemCenter, ItemCenter,
		       ItemCenter, ItemCenter, 0)

    Images[n] = Image.new(normal, bright)
  end

  def build_empty_surface
    surface = SDL::Surface.new(SDL::SWSURFACE, ItemBound, ItemBound, @screen)
    surface.draw_filled_rect_alpha(0, 0, ItemBound, ItemBound,
				 @screen.map_rgb(0, 255, 0), 255)
    surface.set_color_key(SDL::SRCCOLORKEY, surface[0, 0])
    surface
  end

  def build_base_surface(color)
    surface = build_empty_surface
    surface.draw_filled_rect_alpha(@margin, ItemCenter - ItemHeight / 2,
				   ItemWidth, ItemHeight, color, 255)
    surface
  end
end

if __FILE__ == $0
  SDL.init(SDL::INIT_VIDEO)
  screen = SDL.set_video_mode(ScreenWidth, ScreenHeight, 24, SDL::SWSURFACE)

  SDL::WM.set_caption('ristet', 'ristet')

  SDL::TTF.init
  font = SDL::TTF.open("boxfont2.ttf", 24)
  model = Model.new
  view = View.new(model, screen, ScreenWidth / 2 - PanelWidth / 2, ItemCenter,
		  PanelWidth, PanelHeight, font)

  wait = 0.1
  fixed = 0
  loop do
    model.create_item(rand(12))
    finished, paused = false, false
    until model.stopped?
      # handle keystrokes
      while event = SDL::Event2.poll
	case event
	when SDL::Event2::Quit then exit
	when SDL::Event2::KeyDown
	  case event.sym
	  when SDL::Key::ESCAPE, ?q
	    exit
	  when SDL::Key::DOWN
	    finished = true
	    break
	  when SDL::Key::LEFT
	    model.modify_item(model.current_item.y,
			      (model.current_item.angle - 1) % 12)
	  when SDL::Key::RIGHT
	    model.modify_item(model.current_item.y,
			      (model.current_item.angle + 1) % 12)
	  when SDL::Key::PAUSE, ?p
	    paused = !paused
	  end
	end
      end
      unless paused
	model.modify_item(model.current_item.y - 8,
			  model.current_item.angle)
      end
      break if finished

      SDL::Key.scan
      sleep(wait)
    end
    break unless model.fixable?
    model.fix_item
    fixed += 1
    if fixed % 16 == 0
      model.level_up
      wait *= 0.9
    end
  end
  puts("!!! GAME OVER !!!: SCORE #{model.score} (LEVEL #{model.level})")
  sleep(3)
end
